package provide Dialog


# Dialog support library
# ======================
#	- Dialog creation and destruction
#	- Waiting for user action, reactivation
#	- Establishing text panel bindings
#	- Error and warning dialog support



    global dialog_list
		# geo,<dd>	    ... top-level geometry string
		# callback,<dd>	    ... user destroy callback procedure
		# wait,<dd>	    ... wait variable
		# grab,<dd>	    ... preceding grab
		# cursor,<dd>	    ...	save cursor


namespace eval Dialog {


# Dialog panel support procedures
# ===============================


# Check whether a dialog does currently exist
# -------------------------------------------
#
#	Return 1 if it exists, 0 if not

    proc DialogExists { p_name } {
			    # p_name	name of dialog

	global		dialog_list

	return [ info exists dialog_list(geo,$p_name) ]
    }



# Position the dialog relative to widget
# --------------------------------------
#
#	Place top left-hand corner at center of widget

    proc DialogGeometry { p_widget } {
			    # p_widget	reference widget or ""

	if { $p_widget == "" } { 
	    set x_string [ wm geometry . ]
	    regexp {^(\d*)\D(\d*)\D(\d*)\D(\d*)$} $x_string "" \
						x_width x_height x_pos y_pos
	} else { 
	    set x_top [ winfo toplevel $p_widget ]
	    set x_string [ wm geometry $x_top ]
	    regexp {^([0-9\-]*)\D([0-9\-]*)\D([0-9\-]*)\D([0-9\-]*)$} \
				$x_string "" x_width x_height x_pos y_pos
	    set x_widget $p_widget 
	    set x_width [ winfo width $p_widget ]
	    set x_height [ winfo height $p_widget ]
	    while { $x_widget != $x_top } {
		incr x_pos [ winfo x $x_widget ]
		incr y_pos [ winfo y $x_widget ]
	        set x_widget [ winfo parent $x_widget ]
	    }
	}
	set x_pos [ expr ( $x_width / 2 ) + $x_pos + 4 ]
	set y_pos [ expr ( $x_height / 2 ) + $y_pos + 24 ]
	return "+$x_pos+$y_pos"
    }



# Create a new dialog
# -------------------

    proc DialogCreate { p_dialog p_widget p_title { p_callback "" } } {
			    # p_dialog	name of dialot
			    # p_widget	widget triggering this action
			    # p_title	title of new dialog
			    # p_callback user destroy callback procedure

	global		dialog_list

	if [winfo exists $p_dialog] {
	    switch -- [wm state $p_dialog] {
		normal {
				# Raise a buried window
		    raise $p_dialog
		}

		withdrawn -
		    iconic {
				# Open and restore geometry
		    wm deiconify $p_dialog
		    catch  {wm geometry $p_dialog $dialog_list(geo,$p_dialog) }
		}
	    }
	    return 0
	} else {
	    eval { toplevel $p_dialog }
	    wm title $p_dialog $p_title
	    if { $p_widget != "" } {
		set x_geometry [ DialogGeometry $p_widget ]
	    } else { set x_geometry "+0+0" }
	    set dialog_list(geo,$p_dialog) $x_geometry
	    if { $p_callback != "" } { set dialog_list(callback,$p_dialog) \
								$p_callback }
	    catch {wm geometry $p_dialog $dialog_list(geo,$p_dialog)}
	    bind $p_dialog <Destroy> [ list \
				[ namespace current ]::DialogKill $p_dialog ]
	    return 1
	}
    }



# Activate a waiting dialog
# -------------------------

    proc DialogActivate { p_dialog p_value } {
			    # p_dialog	name of dialog to activate
			    # p_value	exit-code after wait

	global		dialog_list

	set dialog_list(wait,$p_dialog) $p_value
    }



# Wait for user action to wake up the dialog
# ------------------------------------------

    proc DialogWait { p_widget p_dialog { p_focus {} } } {
			    # p_widget	widget having launched dialog
			    # p_dialog	name of dialog
			    # p_focus	

	global		dialog_list

	if { [ string length $p_focus ] == 0 } { set p_focus $p_dialog }

	if { $p_widget != "" } {
	    set x_parent [ winfo toplevel $p_widget ]
	} else {
	    set x_parent .
	}
	set save_cursor [ $x_parent cget -cursor ]
	$x_parent config -cursor draped_box
	set old [ focus -displayof $p_dialog ]
	focus $p_focus
	set dialog_list(grab,$p_dialog) [ grab current ]
	catch { grab $p_dialog }

	    #	Wait for the dialog to complete

	set dialog_list(wait,$p_dialog) ""
	tkwait variable dialog_list(wait,$p_dialog)
	if { [ info exists dialog_list(wait,$p_dialog) ] } {
	    catch { $x_parent config -cursor $save_cursor }
	    set exit_code $dialog_list(wait,$p_dialog)
	    unset dialog_list(wait,$p_dialog)
	} else {
	    set exit_code ""
	}
	if { $exit_code == "" } return ""
	catch { grab release $p_dialog }
	if { $dialog_list(grab,$p_dialog) != "" } {
	    catch { grab $dialog_list(grab,$p_dialog) }
	}
	unset dialog_list(grab,$p_dialog)
	focus $old
	return $exit_code
    }



# Kill callback procedure for dialog termination
# ----------------------------------------------

    proc DialogKill { p_dialog } {
			     # p_dialog	dialog being destroyed

	global		dialog_list

	if [ winfo exists $p_dialog ] { bind $p_dialog <Destroy> { } }
	if [ info exists dialog_list(wait,$p_dialog) ] {
	    set dialog_list(wait,$p_dialog) ""
	    update idletasks
	}

	if { [ info exists dialog_list(geo,$p_dialog) ] } {
	    unset dialog_list(geo,$p_dialog)
	}
	if { [ info exists dialog_list(callback,$p_dialog) ] } {
	    eval $dialog_list(callback,$p_dialog)
	    unset dialog_list(callback,$p_dialog)
	}
	if [ info exists dialog_list(save,$p_dialog) ] {
	    unset dialog_list(save,$p_dialog)
	}
	if { $p_dialog == "." } exit
	
    }


    proc DialogBindings { } {

	    # Bindings for normal text panels
	    # --------------------------------

	foreach x_bind [ bind Text ] {
	    if { [ string first Return $x_bind ] >= 0 } {
		bind TextPanelBindings <Return> { }
		continue
	    } elseif { [ string first Tab $x_bind ] >= 0 } {
		bind TextPanelBindings <Tab> { }
		continue
	    }
	    set x_code [ bind Text $x_bind ]
	    bind TextPanelBindings $x_bind $x_code
	}

	    # Bindings for passive text panels
	    # --------------------------------

	foreach x_bind [ bind Text ] {

	    if {    ( [ string first <Button-1	$x_bind ] >= 0 ) || \
		    ( [ string first <B1	$x_bind ] >= 0 ) || \
		    ( [ string first <Key-Left	$x_bind ] >= 0 ) || \
		    ( [ string first <Key-Right	$x_bind ] >= 0 ) || \
		    ( [ string first <Key-Up	$x_bind ] >= 0 ) || \
		    ( [ string first <Key-Down	$x_bind ] >= 0 ) || \
		    ( [ string first <Key-Home	$x_bind ] >= 0 ) || \
		    ( [ string first <Key-End	$x_bind ] >= 0 )	} { 
		set x_code [ bind Text $x_bind ]
		bind PassivePanelBindings $x_bind $x_code
	    } else { 
		bind PassivePanelBindings $x_bind { break }
	    }
	}
    }



# Application interface procedures
# ================================


# Dialog message modal dialog, very similar to tk_dialog
# ------------------------------------------------------
#	p_text ...	list of text lines to figure in the upper part
#			    leading \<nn> :
#				- text is otherwise empty: add vertical spacing
#				- followed by \text: right indent text

    proc DialogMessage { p_parent p_title p_icon p_text p_button } {
			    # p_parent	parent frame
			    # p_title	title text
			    # p_icon	panel header icon
			    # p_text	list of text lines
			    # p_button	list of buttons

	set	MODULE "message"

	set	f_dialog .$MODULE
	if { [ Dialog::DialogCreate $f_dialog $p_parent \
						$p_title ] == 0 } { return 0 }
	wm resizable $f_dialog false false


	    # Top part
	    # --------

	set	f_top	$f_dialog.top
	frame	$f_top
	pack	$f_top -side top -padx 20 -pady 6 -fill x -expand true

	frame	$f_top.space1 -height 12
	pack	$f_top.space1 -side top

	set x_x 0
	foreach x_item $p_text {
	    set x_value ""
	    set x_rest ""
	    if [ regexp {\\(\d+)\\*(.*)} $x_item "" x_value x_text ] {
		if { $x_text == "" } {
		    frame   $f_top.label$x_x -height $x_value
		} else {
		    frame   $f_top.label$x_x
		    frame   $f_top.label${x_x}.a -width $x_value
		    pack    $f_top.label${x_x}.a -side left -anchor w
		    label   $f_top.label${x_x}.b -text $x_text
		    pack    $f_top.label${x_x}.b -side left
		}
	    } else {
		label   $f_top.label$x_x -text $x_item
	    }
	    pack    $f_top.label$x_x -side top -anchor w
	    incr x_x
	}

	frame	$f_dialog.t_sep -relief groove -borderwidth 1 -height 2
	pack	$f_dialog.t_sep -fill x -expand true -side top -padx 0 -pady 0


	    # Bottom part
	    # -----------

	set	f_bottom  $f_dialog.bottom
	frame	$f_bottom
	pack	$f_bottom -side top -padx 0 -pady 6 -fill x -expand true


	set x_x 0
	foreach x_item $p_button {
	    button  $f_bottom.label$x_x -text [ lindex $p_button $x_x ] \
		-highlightthickness 0 -command [ list \
		[ namespace current ]::DialogDone $f_bottom.label$x_x $x_x ]
	    grid    $f_bottom.label$x_x -column $x_x -row 0 -padx 8 -pady 4
	    incr    x_x
	}
	update idletasks
	raise $f_dialog
	set exit_code [ Dialog::DialogWait $p_parent $f_dialog ]
	destroy $f_dialog
	return $exit_code
    }



# Callback for done buttons (ok and ko)
# -------------------------------------

    proc DialogDone { p_widget { p_ok "" } } {
			    # p_widget	activated dialog button

	if { $p_ok == "" } {
	    set p_ok [ $p_widget cget -text ]
	}
	set top [winfo toplevel $p_widget ]
	Dialog::DialogActivate $top $p_ok
    }



# Darken or lighten a colour
# --------------------------

    proc AdjustColour { p_window p_colour p_factor } {
			    # p_widget	widget
			    # p_colour	original colour
			    # p_factor	> 1.0  ... lighten
			    #		< 1.0  ... darken

	set x_rgb [ winfo rgb $p_window $p_colour ]

	for { set ii 0 } { $ii < 3 } { incr ii } {
	    set x_x($ii) [ expr round([lindex $x_rgb $ii] * $p_factor) ]
	    if { $x_x($ii) > 65535 } { set x_x($ii) 65535 }
	}
	return  [ format "#%04x%04x%04x" $x_x(0) $x_x(1) $x_x(2) ]
    }



# Common handling: initialisation, outer frame and control buttons
# ----------------------------------------------------------------
#	Return outer frame widget ("" if dialog already exists)
#		Button-list options:
#		    first character == " " ... add extra padding
#		    last character  == "+" ... button is default

    proc DialogTop { p_dialog p_parent p_buttons p_title { p_nobar "" } } {
			  # p_dialog	dialog name (path)
			  # p_parent	parent widget
			  # p_buttons	list of control button labels
			  # p_title	panel title

	variable	::Config::conf_params

	if { [ Dialog::DialogCreate $p_dialog $p_parent $p_title ] == 0 } \
								{ return "" }
	wm resizable $p_dialog 0 0


	    # Create control part with buttons

	set	f_control $p_dialog.control
	frame	$f_control -padx $conf_params(windowMargin)
	pack	$f_control -side top -fill x
	grid	rowconfigure $f_control 0 -minsize $conf_params(windowMargin)

	set	x_x 0
	foreach	x_label $p_buttons {
	    if { [ string range $x_label 0 0 ] == " " } {
		set x_label [ string range $x_label 1 end ]
		set x_pad 10
	    } else {
		set x_pad 0
	    }
	    if { [ string range $x_label end end ] == "+" } {
		set x_label [ string range $x_label 0 end-1 ]
		set x_default 1
	    } else {
		set x_default 0
	    }
	    set	    x_temp $f_control.button_$x_x
	    set     x_col [ expr $x_x + 2 ]
	    button  $x_temp -text $x_label -command [ list \
						Dialog::DialogDone $x_temp ]
	    if $x_default { $x_temp configure -default active }
	    grid    $x_temp -column $x_col -row 1 -sticky e -padx $x_pad
	    if { [ expr ( [ llength $p_buttons ] - $x_x ) % 2 ] == 0 } {
		grid configure $x_temp -padx $conf_params(buttonSpacing)
	    }
	    grid columnconfigure $f_control $x_col -weight 0
	    incr x_x
	}
	grid	columnconfigure $f_control 0 -weight 0
	grid	columnconfigure $f_control 1 -weight 1

	if { $p_nobar == "" } {
	  frame    $p_dialog.sep1 -relief raised -height 2 -borderwidth 2
	  pack    $p_dialog.sep1 -side top -fill x \
					-pady $conf_params(windowMargin)
	} else {
	    grid rowconfigure $f_control 2 -minsize $conf_params(windowMargin)
	}
	return $f_control
    }



# Launch default button action (if there is a default button)
# -----------------------------------------------------------
#	Default button must be in first child of p_dialog

    proc DialogDefault { p_dialog } {

	set f_control [ lindex [ winfo children $p_dialog ] 0 ]
	foreach x_widget [ winfo children $f_control ] {
	    if { [ winfo class $x_widget ] != "Button" } continue
	    set x_list [ $x_widget configure -default ]
	    if { [ lsearch $x_list "active" ] < 0 } continue
	    set x_list [ $x_widget cget -command ]
	    [ lindex $x_list 0 ] [ lindex $x_list 1 ]
	    return -code break
	}
    }
}
